package ca.uwaterloo.gp.editorlist;

/**********************************************************************
Copyright (c) 2002 IBM Corp.  All rights reserved.
This file is made available under the terms of the Common Public License v1.0
which accompanies this distribution, and is available at
http://www.eclipse.org/legal/cpl-v10.html
*********************************************************************/

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.jface.action.IMenuListener;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.action.Separator;
import org.eclipse.jface.viewers.DecoratingLabelProvider;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IEditorReference;
import org.eclipse.ui.IFileEditorInput;
import org.eclipse.ui.IViewPart;
import org.eclipse.ui.IWorkbenchActionConstants;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.IWorkingSetManager;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.internal.PartSite;
import org.eclipse.ui.part.ViewPart;

/**
 * The EditorList is a view for managing editors. It displays the 
 * currently open editors and allows users to switch editors, close
 * editors, create working sets of editors, and open working sets
 * of editors.
 */
public class EditorList extends ViewPart { /* FRL_05 */
	/**
	 * This view's viewer, which displays the editors as a table
	 */
	private EditorListViewer fViewer;
	private EditorListContentProvider fContentProvider;
	/**
	 * The listeners for the editor list.
	 */
	private EditorListListener fListener;

	/**
	 * Static initializer. Set the default preference values.
	 */
	static {
		PlatformUI.getPreferenceStore().setDefault(IEditorListConstants.PREF_SHOW_RESOURCE_ATTRIBUTES, true);
	}
	
	/**
	 * Returns the currently active instances of the EditorList view in the
	 * workbench.
	 * 
	 * @return EditorList[] the currently active instances of the EditorList
	 * view in the workbench
	 */
	public static EditorList[] getEditorLists() {
		IWorkbenchWindow[] windows= PlatformUI.getWorkbench().getWorkbenchWindows();
		List views= new ArrayList();
		IViewPart view;
		for (int i = 0; i < windows.length; i++) {
			view= windows[i].getActivePage().findView(IEditorListConstants.EDITORLIST_ID);
			if (view != null) {
				views.add(view);
			}
		}
		return (EditorList[]) views.toArray(new EditorList[views.size()]);
	}
	
	/**
	 * Returns the file associated with the given editor part
	 * or <code>null</code> if none could be determined
	 * @param editor the editor
	 * @return the file associated with the given editor or
	 *  <code>null</code>
	 */
	public static IFile getFile(IEditorPart editor) {
		IEditorInput input= editor.getEditorInput();
		if (input instanceof IFileEditorInput) {
			return ((IFileEditorInput) input).getFile();
		}
		return null;
	}

	/**
	 * Returns the IEditorReference wrapping the given IEditorPart
	 * or <code>null</code> if no reference was found.
	 */
	private static IEditorReference getReference(IEditorPart part) {
		return (IEditorReference) ((PartSite) part.getSite()).getPane().getPartReference();
	}

	/**
	 * Creates a new editor list. Defined here to allow for instantiation
	 * via reflection.
	 */
	public EditorList() {
		super();
	}

	/**
	 * @see IWorkbenchPart#createPartControl(Composite)
	 */
	public void createPartControl(Composite parent) { /* FRL_09 */
		fListener = new EditorListListener(this);
		getPage().addPartListener(fListener);
		createViewer(parent);
		createToolbarActions();
		PlatformUI.getWorkbench()/* FRL_10 */.getActiveWorkbenchWindow().addPageListener(fListener);
	}

	/**
	 * Create the view's viewer which displays the open editors
	 */
	private void createViewer(Composite parent) {
		fViewer = new EditorListViewer(parent, SWT.MULTI | SWT.H_SCROLL | SWT.V_SCROLL); /* FRL_14 */
		DecoratingLabelProvider labelProvider= new DecoratingLabelProvider(new EditorListLabelProvider() /* FRL_13 */, 
				PlatformUI.getWorkbench()/* FRL_10 */.getDecoratorManager().getLabelDecorator()); 
		fViewer.setLabelProvider(labelProvider); /* FRL_15 */
		labelProvider.addListener(fListener);
		fViewer.getTable().addMouseListener(fListener);
		fContentProvider= new EditorListContentProvider(); /* FRL_18 */
		fViewer.setContentProvider(fContentProvider); /* FRL_19 */
		fViewer.setInput(getPage()); /* FRL_20 */
		initViewerContents();
		getSite().setSelectionProvider(fViewer);
		fViewer.getControl().addKeyListener(fListener);
	}
	
	/**
	 * Adds the editor references from the active page to the view. It is
	 * important that this method is called at startup because it initializes
	 * the propertyChanged listener for the editor references open at startup
	 */
	private void initViewerContents() {
		IEditorReference[] references= getPage().getEditorReferences();
		for (int i = 0; i < references.length; i++) {
			addReferenceToViewer(references[i]);
		}
	}

	/**
	 * Create the toolbars which appear on the view
	 */
	private void createToolbarActions() {
		IToolBarManager toolbar = getViewSite()/* FRL_30 */.getActionBars()/* FRL_26 */.getToolBarManager()/* FRL_27 */;
		toolbar.add(new SortEditorListAction(fViewer)/* FRL_24 */); /* FRL_29 */
		toolbar.add(new OpenEditorWorkingSetAction()/* FRL_24 */); /* FRL_29 */
	}

	/**
	 * Close all open editors
	 */
	public void closeAllEditors() {
		IWorkbenchPage page = getPage();
		page.closeAllEditors(true);
		refreshContents(); // Workaround to lack of IEditorReference.close() notifications
	}

	/**
	 * Close the editors that are currently selected in the viewer
	 */
	public void closeSelectedEditors() {
		StructuredSelection selection = (StructuredSelection) fViewer.getSelection();
		IWorkbenchPage page = getPage();
		Iterator iter = selection.iterator();
		EditorListElement element;
		while (iter.hasNext()) {
			element = (EditorListElement) iter.next();
			IEditorPart editor = getRestoredEditor(element.getEditorReference());
			if (editor != null) {
				page.closeEditor(editor, true);
			}
		}
		refreshContents(); // Workaround to lack of IEditorReference.close() notifications
	}

	/**
	 * Creates a new working set with the given name composed of
	 * the currently selected editors.
	 * 
	 * @param name the name of the working set to be created.
	 */
	public void createWorkingSetFromSelectedEditors(String name) {
		StructuredSelection selection = (StructuredSelection) fViewer.getSelection();
		Iterator iter = selection.iterator();
		List list = new ArrayList();
		EditorListElement element;
		IEditorPart editor;
		while (iter.hasNext()) {
			element = (EditorListElement) iter.next();
			editor = getRestoredEditor(element.getEditorReference());
			if (editor != null) {
				IFile file= getFile(editor);
				if (file != null) {
					list.add(file);
				}
			}
		}
		if (list.size() == 0) {
			return;
		}
		IAdaptable adaptables[] = new IAdaptable[list.size()];
		list.toArray(adaptables);
		IWorkingSetManager manager = PlatformUI.getWorkbench().getWorkingSetManager();
		manager.addWorkingSet(manager.createWorkingSet(name, adaptables));
	}

	/**
	 * Attempts to get the restored editor part for the given reference.
	 * If the part fails to restore, the given reference is removed from the viewer.
	 */
	protected IEditorPart getRestoredEditor(IEditorReference reference) {
		IEditorPart part = reference.getEditor(true);
		if (part == null) {
			removeEditorReference(reference);
		}
		return part;
	}

	/**
	 * Returns the workbench page that contains this view
	 */
	protected IWorkbenchPage getPage() {
		return getSite().getPage();
	}
	
	/**
	 * Returns the editor that is selected in the view or <code>null</code>
	 * if no editor is selected.
	 */
	protected IEditorReference getSelectedEditor() {
		ISelectionProvider provider= getViewSite().getSelectionProvider();
		if (provider == null) {
			return null;
		}
		IStructuredSelection selection= (IStructuredSelection)provider.getSelection();
		if (selection.size() != 1) {
			return null;
		}
		return ((EditorListElement) selection.getFirstElement()).getEditorReference();
	}

	/**
	 * Notifes this view that the given editor part has closed
	 */
	public void editorClosed(IEditorPart editor) {
		removeEditorReference(getReference(editor));
	}

	/**
	 * Removes the given editor reference from the view
	 */
	public void removeEditorReference(IEditorReference reference) {
		reference.removePropertyListener(fListener);
		fViewer.remove(fContentProvider.getElement(reference));
		fContentProvider.removeElement(reference);
	}

	/**
	 * Adds the given editor part to the view. If the given editor
	 * is already represented in the view, this method has no effect.
	 */
	public void addEditorToViewer(IEditorPart editor) {
		TableItem[] items = fViewer.getTable().getItems();
		EditorListElement element;
		IEditorReference reference= getReference(editor);
		// Ensure that the reference for this editor part hasn't already
		// been added.
		for (int i = 0, numItems = items.length; i < numItems; i++) {
			element = (EditorListElement)items[i].getData();
			if (element.getEditorReference() == reference) {
				return;
			}
		}
		addReferenceToViewer(reference);
	}
		
	/**
	 * Adds the given editor reference to the viewer even if the given editor
	 * is already represented in the view.
	 */
	public void addReferenceToViewer(IEditorReference reference) {
		reference.addPropertyListener(fListener);
		EditorListElement element= new EditorListElement(reference);
		element= fContentProvider.addElement(reference);
		fViewer.add(element);
	}

	/**
	 * Refreshes the contents of this view's viewer.
	 */
	public void refreshContents() {
		fViewer.getTable().removeAll();
		fViewer.refresh();
	}

	/**
	 * Refreshes the label of the editor element associated with the given
	 * editor part.
	 * 
	 * @param part the part whose editor element label needs to be refreshed
	 */
	public void refreshLabel(IEditorPart part) {
		fViewer.refresh(fContentProvider.getElement(getReference(part)));
	}
	
	/**
	 * Activates the editor that is currently selected. This makes the selected
	 * editor visible and gives the editor focus.
	 */
	public void activateSelectedEditor() {
		IEditorReference reference= getSelectedEditor();
		if (reference == null) {
			return;
		}
		IEditorPart editor = getRestoredEditor(reference);
		getPage().activate(editor);
		updateStatusLineFor(editor);
	}
	
	/**
	 * Brings to top the editor that is currently selected. This makes the
	 * selected editor visible and keeps focus in the view.
	 */
	public void bringToTopSelectedEditor() {
		IEditorReference reference= getSelectedEditor();
		if (reference == null) {
			return;
		}
		IEditorPart editor = getRestoredEditor(reference);
		if (editor != getPage().getActiveEditor()) {
			getPage().bringToTop(editor);
			updateStatusLineFor(editor);
		}
	}
	
	/**
	 * Updates the status line to display the full path to the file
	 * associated with the given editor if such a file exists.
	 * @param editor the editor to update for
	 */
	public void updateStatusLineFor(IEditorPart editor) {
		IFile file= getFile(editor);
		String message= null;
		if (file != null) {
			message= file.getFullPath().toString();
		}
		getViewSite().getActionBars().getStatusLineManager().setMessage(message);
	}

	/**
	 * The given editor has been selected. Update the
	 * viewer selection.
	 */
	public void editorSelected(IEditorPart editor) {
		IEditorReference selectedEditor= getSelectedEditor();
		IEditorReference newEditor= getReference(editor);
		if (selectedEditor != null) {
			if (selectedEditor.equals(newEditor)) {
				return;
			}
		}
		EditorListElement element= fContentProvider.getElement(newEditor);
		if (element != null) {
			fViewer.setSelection(new StructuredSelection(element), true);
		}
	}

	/**
	 * @see IWorkbenchPart#setFocus()
	 */
	public void setFocus() { /* FRL_06 */
		fViewer.getControl().setFocus(); /* FRL_07 */ /* FRL_08 */
	}

	/**
	 * @see IWorkbenchPart#dispose()
	 */
	public void dispose() {
		super.dispose();
		getPage().removePartListener(fListener);
		IEditorReference[] references = getPage().getEditorReferences();
		for (int i = 0, numReferences = references.length; i < numReferences; i++) {
			references[i].removePropertyListener(fListener);
		}
		PlatformUI.getWorkbench().getActiveWorkbenchWindow().removePageListener(fListener);
	}
}
